#!/usr/bin/perl
# UseModWiki version 0.88 (October 7, 2000)
# Copyright (C) 2000 Clifford A. Adams
#    <caadams@frontiernet.net> or <usemod@usemod.com>
# Based on the GPLed AtisWiki 0.3  (C) 1998 Markus Denker
#    <marcus@ira.uka.de>
# ...which was based on
#    the LGPLed CVWiki CVS-patches (C) 1997 Peter Merel
#    and The Original WikiWikiWeb  (C) Ward Cunningham
#        <ward@c2.com> (code reused with permission)
# Email and ThinLine options by Jim Mahoney <mahoney@marlboro.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the
#    Free Software Foundation, Inc.
#    59 Temple Place, Suite 330
#    Boston, MA 02111-1307 USA

$| = 1;    # Do not buffer output
undef $/;  # Read complete files
use strict;
# Configuration/constant variables:
use vars qw(@RcDays @HtmlPairs @HtmlSingle $LockDir $DataDir $HtmlDir
  $UserDir $InterFile $RcFile $FullUrl $SiteName $HomePage $ScriptTZ
  $LogoUrl $RcDefault $IndentLimit $RecentTop $EditAllowed $UseDiff
  $UseSubpage $UseCache $RawHtml $SimpleLinks $NonEnglish $LogoLeft
  $HtmlTags $HtmlLinks
  $UseEmailNotify $SendMail $NotifyDefault $EmailFrom
  $ShowEdits $ThinLine $LinkPattern $InterLinkPattern $InterSitePattern
  $UrlProtocols $UrlPattern $ImageExtensions $RFCPattern $ISBNPattern
  $FS $FS1 $FS2 $FS3 $CookieName $SiteBase);
# Other global variables:
use vars qw(%page %FullPage %InterSite %SaveUrl %SaveNumUrl %UserCookie
  %SetCookie %UserData @HtmlStack $InterSiteInit $PageHtml
  $SaveHtmlStack $SaveUrlIndex $SaveNumUrlIndex $MainPage $OpenId
  $q $UserID $TimeZoneOffset $ScriptName $BrowseCode $OtherCode);

# == Configuration =====================================================
$DataDir     = "/tmp/mywikidb"; # Main page database
$CookieName  = "Wiki";          # Name for this wiki (for multi-wiki sites)
$SiteName    = "Wiki";          # Name of site (used for titles)
$HomePage    = "HomePage";      # Home page (must be valid LinkPattern)
$LogoUrl     = "/wiki.gif";     # URL for site logo ("" for no logo)
$ENV{PATH}   = "/usr/bin/";     # Path used to find "diff"
$ScriptTZ    = "";              # Local time zone ("" means do not print)
$RcDefault   = 30;              # Default number of RecentChanges days
@RcDays      = qw(1 3 7 30 90); # Days for links on RecentChanges
$SiteBase    = "";              # Full URL for <BASE> header
$FullUrl     = "";              # Set if the auto-detected URL is wrong

$UseSubpage  = 1;       # 1 = use subpages,       0 = do not use subpages
$UseCache    = 0;       # 1 = cache HTML pages,   0 = generate every page
$EditAllowed = 1;       # 1 = editing allowed,    0 = read-only
$LogoLeft    = 0;       # 1 = logo on left,       0 = logo on right
$RecentTop   = 1;       # 1 = recent on top,      0 = recent on bottom
$UseDiff     = 1;       # 1 = use diff features,  0 = do not use diff
$ShowEdits   = 0;       # 1 = show minor edits,   0 = hide edits by default
$RawHtml     = 0;       # 1 = allow <HTML> tag,   0 = no raw HTML in pages
$HtmlTags    = 0;       # 1 = "unsafe" HTML tags, 0 = only minimal tags
$HtmlLinks   = 0;       # 1 = allow A HREF links, 0 = no raw HTML links
$SimpleLinks = 0;       # 1 = only letters,       0 = allow _ and numbers
$NonEnglish  = 0;       # 1 = extra link chars,   0 = only A-Za-z chars
$ThinLine    = 0;       # 1 = fancy <hr> tags,    0 = classic wiki <hr>

$UseEmailNotify = 0;       # 1 = allow email notification,     0 = don't
$NotifyDefault  = 0;       # Default for email notify checkbox on Edit page.
$EmailFrom      = "Wiki";  # Text for "From: " field of email notes.
                           # (But Reply-to; is set to same as To: anyway.)
$SendMail       = "/usr/sbin/sendmail"; # Full path to sendmail executable

# HTML tag lists, enabled if $HtmlTags is set.
# Scripting is currently possible with these tags,
# so they are *not* particularly "safe".
# Tags that must be in <tag> ... </tag> pairs:
@HtmlPairs = qw(b i u font big small sub sup h1 h2 h3 h4 h5 h6 cite code
  em s strike strong tt var div center blockquote ol ul dl table caption);
# Single tags (that do not require a closing /tag)
@HtmlSingle = qw(br p hr li dt dd tr td th);
@HtmlPairs = (@HtmlPairs, @HtmlSingle);  # All singles can also be pairs

# == You should not have to change anything below this line. =============
$IndentLimit = 20;                  # Maximum depth of nested lists
$LockDir     = "$DataDir/lock";     # DB is locked if this exists
$HtmlDir     = "$DataDir/html";     # Stores HTML versions
$UserDir     = "$DataDir/user";     # Stores user data
$InterFile   = "$DataDir/intermap"; # Interwiki site->url map
$RcFile      = "$DataDir/rclog";    # New RecentChanges logfile

# The "main" program, called at the end of this script file.
sub DoWikiRequest {
  &InitLinkPatterns();
  if (!&DoCacheBrowse()) {
    eval $BrowseCode;
    if (!&DoBrowseRequest()) {
      eval $OtherCode;
      &DoOtherRequest();
    }
  }
}

# == Common and cache-browsing code ====================================
sub InitLinkPatterns {
  my ($UpperLetter, $LowerLetter, $AnyLetter, $LpA, $LpB, $QDelim);

  $UpperLetter = "[A-Z";
  $LowerLetter = "[a-z";
  $AnyLetter   = "[A-Za-z";

  if ($NonEnglish) {
    $UpperLetter .= "\xc0-\xde";
    $LowerLetter .= "\xdf-\xff";
    $AnyLetter   .= "\xc0-\xff";
  }
  if (!$SimpleLinks) {
    $AnyLetter .= "_0-9";
  }
  $UpperLetter .= "]"; $LowerLetter .= "]"; $AnyLetter .= "]";

  # Main link pattern: lowercase between uppercase, then anything
  $LpA = $UpperLetter . "+" . $LowerLetter . "+" . $UpperLetter
         . $AnyLetter . "*";
  # Optional subpage link pattern: uppercase, lowercase, then anything
  $LpB = $UpperLetter . "+" . $LowerLetter . "+" . $AnyLetter . "*";

  if ($UseSubpage) {
    # Loose pattern: If subpage is used, subpage may be simple name
    $LinkPattern = "((($LpA)?\\/$LpB)|$LpA)";
    # Strict pattern: both sides must be the main LinkPattern
    # $LinkPattern = "((($LpA)?\\/)?$LpA)";
  } else {
    $LinkPattern = "($LpA)";
  }
  $QDelim = '("")?';     # Optional quote delimiter (not in output)
  $LinkPattern .= $QDelim;

  # Url-style links are delimited by one of:
  #   1.  Whitespace (kept in output)
  #   2.  Left angle-bracket (<)  (kept in output)
  #   3.  A single double-quote (")  (kept in output)
  #   4.  A double double-quote ("") (removed from output)

  # Inter-site convention: sites must start with uppercase letter
  # (Uppercase letter avoids confusion with URLs)
  $InterSitePattern = $UpperLetter . $AnyLetter . "+";
  $InterLinkPattern = "(($InterSitePattern:[^\\s\"<]+)$QDelim)";
  $UrlProtocols = "(http|ftp|afs|news|nntp|mid|cid|mailto|wais|"
                  . "prospero|telnet|gopher)";
  $UrlPattern = "((($UrlProtocols):[^\\s\"<]+)$QDelim)";
  $ImageExtensions = "(gif|jpg|png|bmp|jpeg)";
  $RFCPattern = "RFC\\s?(\\d+)";
  $ISBNPattern = "ISBN:?([0-9- xX]{10,})";
}

sub GetPageDirectory {
  my ($id) = @_;

  if ($id =~ /^([a-zA-Z])/) {
    return uc($1) . "/";
  }
  return "other/";
}

# Simple HTML cache
sub DoCacheBrowse {
  my ($query, $id_file, $text);

  return 0  if (!$UseCache);

  $query = $ENV{'QUERY_STRING'};
  if (($query eq "") && ($ENV{'REQUEST_METHOD'} eq "GET")) {
    $query = $HomePage;  # Allow caching of home page.
  }
  if (!($query =~ /^$LinkPattern$/)) {
    return 0;  # Only use cache for simple links
  }
  $id_file = $HtmlDir . "/" . &GetPageDirectory($query) . "$query.htm";
  if (-f "$id_file") {
    open(INFILE, "<$id_file") or return 0;
    $text = <INFILE>;
    close INFILE;
    print $text;
    return 1;
  }
  return 0;
}

# == Normal page-browsing and RecentChanges code =======================
$BrowseCode = ""; # Comment next line to always compile (slower)
$BrowseCode = <<'#END_OF_BROWSE_CODE';
use CGI;
use CGI::Carp qw(fatalsToBrowser);
$CGI::POST_MAX=1024 * 200;  # max 200K posts
$CGI::DISABLE_UPLOADS = 1;  # no uploads
$q = new CGI;

sub DoBrowseRequest {
  my ($id, $action, $text);

  &InitRequest() or return 1;
  &InitCookie();
  if (!$q->param) {             # No parameter
    &BrowsePage($HomePage);
    return 1;
  }
  $id = &GetParam("keywords", "");
  if ($id) {                    # Just script?PageName
    &BrowsePage($id)  if &ValidId($id);
    return 1;
  }
  $action = lc(&GetParam("action", ""));
  $id = &GetParam("id", "");
  if ($action eq "browse") {
    &BrowsePage($id)  if &ValidId($id);
    return 1;
  } elsif ($action eq "rc") {
    &BrowsePage("RecentChanges");
    return 1;
  }
  return 0;  # Request not handled
}

sub InitRequest {
  my @ScriptPath = split('/', "$ENV{SCRIPT_NAME}");

  $ScriptName = pop(@ScriptPath);  # Name used in links
  # The FS character is a superscript "3"
  $FS = "\xb3"; $FS1 = $FS . "1"; $FS2 = $FS . "2"; $FS3 = $FS . "3";
  $MainPage = ".";  # Directory of the main page (used for subpages)
  $InterSiteInit = 0;
  %InterSite = ();
  $OpenId = "";
  if (!chdir $DataDir) {
    mkdir($DataDir,0775)  if (!(-d $DataDir));
    if (!chdir $DataDir) {
      &ReportError("Could not go to or create $DataDir: $!");
      return 0;
    }
  }
  return 1;
}

sub InitCookie {
  %SetCookie = ();
  $TimeZoneOffset = 0;
  %UserCookie = $q->cookie($CookieName);
  $UserID = $UserCookie{'id'};
  $UserID =~ s/\D//g;  # Numeric only
  if ($UserID < 1) {
    $UserID = 111;
  } else {
    &LoadUserData($UserID);
  }
  if ($UserID > 199) {
    if (($UserData{'id'}       != $UserCookie{'id'})      ||
        ($UserData{'randkey'}  != $UserCookie{'randkey'}) ||
        ($UserData{'password'} != $UserCookie{'password'})) {
      $UserID = 113;
      %UserData = ();   # Invalid.  Later consider warning message.
    }
  }
  if ($UserData{'tzoffset'} != 0) {
    $TimeZoneOffset = $UserData{'tzoffset'} * (60 * 60);
  }
}

sub LoadUserData {
  my ($userFile, $data);

  %UserData = ();
  $userFile = &UserDataFilename($UserID);
  if (open(IN, "<$userFile")) {
    $data=<IN>;
    close IN;
  } else {
    $UserID = 112;  # Could not open file.  Later warning message?
    return;
  }
  %UserData = split(/$FS1/, $data, -1);  # -1 keeps trailing null fields
}

sub UserDataFilename {
  my ($id) = @_;

  return ""  if ($id < 1);
  return $UserDir . "/" . ($id % 10) . "/$id.db";
}

sub GetParam {
  my ($name, $default) = @_;
  my $result;

  $result = $q->param($name);
  if (!defined($result)) {
    if (defined($UserData{$name})) {
      $result = $UserData{$name};
    } else {
      $result = $default;
    }
  }
  return $result;
}

sub BrowsePage {
  my ($id) = @_;
  my ($fullHtml, $oldId, $allDiff, $showDiff);

  %page = &GetCurrentPage($id);
  # Handle a single-level redirect
  $oldId    = &GetParam("oldid", "");
  if (($oldId eq "") && (substr($page{text}, 0, 10) eq "#REDIRECT ")) {
    $oldId = $id;
    ($id) = ($page{text} =~ /\#REDIRECT\s+(\S+)/);
    return  if (!&ValidId($id));
    &ReBrowsePage($id, $oldId);
    return;
  }
  $MainPage = $id;
  $MainPage =~ s|/.*||;  # Only the main page name (remove subpage)

  $fullHtml = &GetHeader($id, &QuoteHtml($id), $oldId);

  $allDiff  = &GetParam("alldiff", 0);
  if ($allDiff != 0) {
    $allDiff = &GetParam("defaultdiff", 1);
  }
  if (($id eq "RecentChanges") && &GetParam("norcdiff", 1)) {
    $allDiff = 0;  # Only show if specifically requested
  }
  $showDiff = &GetParam("diff", $allDiff);
  if ($showDiff) {
    $fullHtml .= &GetDiffHTML($showDiff, $id);
  }
  $fullHtml .= &WikiToHTML($page{text}) . "<hr>\n";
  if ($id eq "RecentChanges") {
    print $fullHtml;
    &ActionRc();
    print "<hr>\n", &GetFooterText($id);
    return;
  }
  $fullHtml .= &GetFooterText($id);
  print $fullHtml;
  return  if ($showDiff);  # No cache for diff version
  &UpdateHtmlCache($id, $fullHtml)  if $UseCache;
}

sub ReBrowsePage {
  my ($id, $oldId) = @_;

  if ($oldId ne "") {   # Target of #REDIRECT (loop breaking)
    $id = "action=browse&id=$id&oldid=$oldId";
  }
  print &GetRedirectPage($id);
}

sub ScriptLink {
  my ($action, $text) = @_;

  return "<a href=\"$ScriptName?$action\">$text</a>";
}

sub ScriptLinkDiff {
  my ($diff, $id, $text) = @_;

  $diff = &GetParam("defaultdiff", 1)  if ($diff == 4);
  return &ScriptLink("action=browse&diff=$diff&id=$id", $text);
}

sub ScriptLinkTitle {
  my ($action, $text, $title) = @_;

  return "<a href=\"$ScriptName?$action\" title=\"$title\">$text</a>";
}

sub GetDiffHTML {
  my ($diffType, $id) = @_;
  my ($html, $diffText, $PriorName, $links, $usecomma);
  my ($major, $minor, $author, $usemajor, $useminor, $useauthor);

  $links = "(";
  $usecomma = 0;
  $major  = &ScriptLinkDiff(1, $id, "major diff");
  $minor  = &ScriptLinkDiff(2, $id, "minor diff");
  $author = &ScriptLinkDiff(3, $id, "author diff");
  $usemajor  = 0;
  $useminor  = 0;
  $useauthor = 0;
  if ($diffType == 1) {
    $diffText = $page{majordiff};
    if ($diffText eq "1") {
      $diffText = $page{minordiff};
    }
    $PriorName = "major version";
    if ($page{majordiff} ne "1") {
      $useminor = 1;
    }
    if ($page{majordiff} ne $page{authordiff}) {
      $useauthor = 1;
    }
  } elsif ($diffType == 2) {
    $diffText = $page{minordiff};
    $PriorName = "minor version";
    if ($page{majordiff} ne "1") {
      $usemajor = 1;
    }
    if ($page{authordiff} ne "1") {
      $useauthor = 1;
    }
  } else {
    $diffText = $page{authordiff};
    if ($diffText eq "1") {
      $diffText = $page{minordiff};
    }
    $PriorName = "author";
    if ($page{authordiff} ne "1") {
      $useminor = 1;
    }
    if ($page{authordiff} ne $page{majordiff}) {
      $usemajor = 1;
    }
  }
  if ($usemajor) {
    $links .= $major;
    $usecomma = 1;
  }
  if ($useminor) {
    $links .= ", "  if ($usecomma);
    $links .= $minor;
    $usecomma = 1;
  }
  if ($useauthor) {
    $links .= ", "  if ($usecomma);
    $links .= $author;
  }
  if (!($usemajor || $useminor || $useauthor)) {
    $links .= "no other diffs";
  }
  $links .= ")";

  $html = "<b>Difference (from prior $PriorName)</b>\n"
          . "$links<br>" . &DiffToHTML($diffText) . "<hr>\n";
  return $html;
}

sub DiffToHTML {
  my ($html) = @_;

  $html =~ s/\n--+//g;
  # Note: Need spaces before <br> to be different from diff section.
  $html =~ s/(^|\n)(\d+.*c.*)/$1 <br><strong>Changed: $2<\/strong><br>/g;
  $html =~ s/(^|\n)(\d+.*d.*)/$1 <br><strong>Removed: $2<\/strong><br>/g;
  $html =~ s/(^|\n)(\d+.*a.*)/$1 <br><strong>Added: $2<\/strong><br>/g;
  $html =~ s/\n((<.*\n)+)/&ColorDiff($1,"ffffaf")/ge;
  $html =~ s/\n((>.*\n)+)/&ColorDiff($1,"cfffcf")/ge;
  return $html;
}

sub ColorDiff {
  my ($diff, $color) = @_;

  $diff =~ s/(^|\n)[<>]/$1/g;
  $diff = &QuoteHtml($diff);
  # Do some of the Wiki markup rules:
  %SaveUrl = ();
  %SaveNumUrl = ();
  $SaveUrlIndex = 0;
  $SaveNumUrlIndex = 0;
  $diff =~ s/$FS//g;
  $diff =~ s/\&lt;nowiki\&gt;((.|\n)*?)\&lt;\/nowiki\&gt;/&StoreRaw($1)/ige;
  $diff =  &CommonMarkup($diff, 0, 1);  # No images, all patterns
  $diff =~ s/$FS(\d+)$FS/$SaveUrl{$1}/geo;
  $diff =~ s/\r?\n/<br>/g;
  return "<table width=\"95\%\" bgcolor=#$color><tr><td>\n" . $diff
         . "</td></tr></table>\n";
}

sub CommonMarkup {
  my ($text, $useImage, $doLines) = @_;
  local $_ = $text;

  if ($doLines < 2) { # 2 = do line-oriented only
    if ($HtmlTags) {
      my ($t);
      foreach $t (@HtmlPairs) {
        s/\&lt;$t(\s[^<>]+?)?\&gt;(.*?)\&lt;\/$t\&gt;/<$t$1>$2<\/$t>/gis;
      }
      foreach $t (@HtmlSingle) {
        s/\&lt;$t(\s[^<>]+?)?\&gt;/<$t$1>/gi;
      }
    } else {
      # Note that these tags are restricted to a single line
      s/\&lt;b\&gt;(.*?)\&lt;\/b\&gt;/<b>$1<\/b>/gi;
      s/\&lt;i\&gt;(.*?)\&lt;\/i\&gt;/<i>$1<\/i>/gi;
    }
    if ($HtmlLinks) {
      s/\&lt;A(\s[^<>]+?)\&gt;(.*?)\&lt;\/a\&gt;/&StoreHref($1, $2)/gise;
    }
    s/\[$UrlPattern\]/&StoreBracketUrl($1)/geo;
    s/\[$InterLinkPattern\]/&StoreBracketInterPage($1)/geo;
    s/$UrlPattern/&StoreUrl($1, $useImage)/geo;
    s/$InterLinkPattern/&StoreInterPage($1)/geo;
    s/$LinkPattern/&GetPageOrEditLink($1)/geo;
    s/$RFCPattern/&StoreRFC($1)/geo;
    s/$ISBNPattern/&StoreISBN($1)/geo;
    if ($ThinLine) {
      s/----+/<hr noshade size=1>/g;
      s/====+/<hr noshade size=2>/g;
    } else {
      s/----+/<hr>/g;
    }
  }
  if ($doLines) { # 0 = no line-oriented, 1 or 2 = do line-oriented
    # The quote markup patterns avoid overlapping tags (with 5 quotes)
    # by matching the inner quotes for the strong pattern.
    s/('*)'''(.*?)'''/$1<strong>$2<\/strong>/g;
    s/''(.*?)''/<em>$1<\/em>/g;
  }
  return $_;
}

sub QuoteHtml {
  my ($html) = @_;

  $html =~ s/&/&amp;/g;
  $html =~ s/</&lt;/g;
  $html =~ s/>/&gt;/g;
  return $html;
}

sub StoreInterPage {
  my ($id) = @_;
  my ($link, $extra);

  ($link, $extra) = &InterPageLink($id);
  # Next line ensures no empty links are stored
  $link = &StoreRaw($link)  if ($link ne "");
  return $link . $extra;
}

sub InterPageLink {
  my ($id) = @_;
  my ($name, $site, $remotePage, $url, $punct);

  ($id, $punct) = &SplitUrlPunct($id);

  $name = $id;
  ($site, $remotePage) = split(/:/, $id, 2);
  $url = &GetSiteUrl($site);
  return ("", $id . $punct)  if ($url eq "");
  $remotePage =~ s/&amp;/&/g;  # Unquote common URL HTML
  $url .= $remotePage;
  return ("<a href=\"$url\">$name</a>", $punct);
}

sub StoreBracketInterPage {
  my ($id) = @_;
  my ($site, $remotePage, $url, $index);

  ($site, $remotePage) = split(/:/, $id, 2);
  $url = &GetSiteUrl($site);
  return "[$id]"  if ($url eq "");
  $index = &GetBracketUrlIndex($id);
  $remotePage =~ s/&amp;/&/g;  # Unquote common URL HTML
  $url .= $remotePage;
  return &StoreRaw("<a href=\"$url\">[$index]</a>");
}

sub GetBracketUrlIndex {
  my ($id) = @_;
  my ($index, $key);

  # Consider plain array?
  if ($SaveNumUrl{$id} > 0) {
    return $SaveNumUrl{$id};
  }
  $SaveNumUrlIndex++;  # Start with 1
  $SaveNumUrl{$id} = $SaveNumUrlIndex;
  return $SaveNumUrlIndex;
}

sub GetSiteUrl {
  my ($site) = @_;
  my $data;
  my $url = "";

  if (!$InterSiteInit) {
    $InterSiteInit = 1;
    open(IN, "<$InterFile") or return "";  # Later warn?
    $data=<IN>;
    close IN;
    %InterSite = split(/\s+/, $data);  # Later consider defensive code
  }
  $url = $InterSite{$site}  if (defined($InterSite{$site}));
  return $url;
}

sub GetPageLink {
  my ($id) = @_;
  my $name = $id;

  $id =~ s|^/|$MainPage/|;
  return &ScriptLink($id, $name);
}

sub GetEditLink {
  my ($id, $linkname) = @_;

  return &ScriptLink("action=Edit&id=$id", $linkname);
}

sub GetPageOrEditLink {
  my ($id) = @_;
  my $name = $id;

  $id =~ s|^/|$MainPage/|;
  if (-f (&GetPageDirectory($id) . "$id.db")) {  # Page exists
    return &GetPageLink($name);
  }
  return $name . &GetEditLink($id,"?");
}

sub GetSearchLink {
  my ($id) = @_;
  my $name = $id;

  $id =~ s|.+/|/|;  # Subpage match: search for just /SubName
  return &ScriptLink("search=$id", $name);
}

sub GetPrefsLink {
  return &ScriptLink("action=editprefs", "Preferences");
}

sub GetRandomLink {
  return &ScriptLink("action=random", "Random Page");
}

sub GetHeader {
  my ($id, $title, $oldId) = @_;
  my $header = "";
  my $LogoImage = "";
  my $result = "";
  my $cookie;

  if ($LogoUrl ne "") {
    $LogoImage = "img src=\"$LogoUrl\" alt=\"[Home]\" border=0";
    if (!$LogoLeft) {
      $LogoImage .= " align=\"right\"";
    }
    $header = &ScriptLink($HomePage, "<$LogoImage>");
  }
  if (defined($SetCookie{'id'})) {
    $cookie = "$CookieName="
            . "rev&" . $SetCookie{'rev'}
            . "&id&" . $SetCookie{'id'}
            . "&randkey&" . $SetCookie{'randkey'};
    $cookie .= ";expires=Fri, 08-Sep-2010 19:48:23 GMT";
    $result = $q->header(-cookie=>$cookie);
  } else {
    $result = $q->header();
  }

  if ($SiteBase ne "") {
    $result .= $q->start_html('-title' => "$SiteName: $title",
                              '-xbase' => $SiteBase,
                              '-BGCOLOR' => 'white');
  } else {
    $result .= $q->start_html('-title' => "$SiteName: $title",
                              '-BGCOLOR' => 'white');
  }
  if ($oldId ne "") {
    $result .= $q->h3("(redirected from "
                      . &GetEditLink($oldId, $oldId) . ")");
  }
  if ($id ne "") {
    $result .= $q->h1($header . &GetSearchLink($id));
  } else {
    $result .= $q->h1($header . $title);
  }
  if (&GetParam("toplinkbar", 1)) {
    # Later consider smaller size?
    $result .= &GetGotoBar($id) . "<hr>";
  }
  return $result;
}

sub GetFooterText {
  my ($id) = @_;
  my $result = "";

  $result = &GetFormStart();
  $result .= &GetGotoBar($id);
  if ($EditAllowed) {
    $result .= &GetEditLink($id, "Edit text of this page");
  } else {
    $result .= "This page is read-only";
  }
  if ($page{timestamp} ne "") {
    $result .= " (last edited " . &TimeToText($page{timestamp});
    if ($UseDiff) {
      $result .= " " . &ScriptLinkDiff(4, $id, "(diff)");
    }
    $result .= ")";
  }
  $result .= "<br>" . &GetSearchForm();
  if ($DataDir =~ m|/tmp/|) {
    $result .= "<br><b>Warning:</b> Database is stored in temporary"
               . " directory $DataDir<br>";
  }
  $result .= $q->endform;
  $result .= $q->end_html;
  return $result;
}

sub GetFormStart {
  return $q->startform("POST", "$ScriptName",
                       "application/x-www-form-urlencoded");
}

sub GetGotoBar {
  my ($id) = @_;
  my ($main, $bartext);

  $bartext  = &GetPageLink($HomePage);
  if ($id =~ m|/|) {
    $main = $id;
    $main =~ s|/.*||;  # Only the main page name (remove subpage)
    $bartext .= " | " . &GetPageLink($main);
  }
  $bartext .= " | " . &GetPageLink("RecentChanges");
  $bartext .= " | " . &GetPrefsLink();
  if (&GetParam("linkrandom", 0)) {
    $bartext .= " | " . &GetRandomLink();
  }
  $bartext .= "<br>\n";
  return $bartext;
}

sub GetSearchForm {
  my ($result);

  $result = "Search: " . $q->textfield(-name=>'search', -size=>20)
            . &GetHiddenValue("dosearch", 1);
  return $result;
}

sub GetHiddenValue {
  my ($name, $value) = @_;

  $q->param($name, $value);
  return $q->hidden($name);
}

sub StackHtml {
  my ($code, $depth) = @_;

  return  if ($depth < 1);  # Protect from bad depth
  if ($depth > $IndentLimit) {
    $depth = $IndentLimit;
  }
  $SaveHtmlStack = 1;
  while (@HtmlStack > $depth) {
    $PageHtml .=  "</" . pop(@HtmlStack) . ">\n";
  }
  while (@HtmlStack < $depth) {
    push(@HtmlStack, $code);
    $PageHtml .= "<$code>\n";
  }
  my $oldCode = pop(@HtmlStack);
  if ($oldCode ne $code) {
    $PageHtml .= "</$oldCode><$code>\n";
  }
  push(@HtmlStack, $code);
}

sub UnstackHtml {
  while (@HtmlStack > 0) {
    $PageHtml .=  "</" . pop(@HtmlStack) . ">\n";
  }
}

sub WikiToHTML {
  my ($pageText) = @_;
  my ($inPre);

  $inPre = 0;
  %SaveUrl = ();
  %SaveNumUrl = ();
  $SaveUrlIndex = 0;
  $SaveNumUrlIndex = 0;
  $PageHtml = "";
  $pageText =~ s/$FS//g;              # Remove separators (paranoia)
  if ($RawHtml) {
    $pageText =~ s/<html>((.|\n)*?)<\/html>/&StoreRaw($1)/ige;
  }
  $pageText = &QuoteHtml($pageText);
  # The <nowiki> tag stores text with no markup (except quoting HTML)
  $pageText =~
    s/\&lt;nowiki\&gt;((.|\n)*?)\&lt;\/nowiki\&gt;/&StoreRaw($1)/ige;
  # The <pre> tag wraps the stored text with the HTML <pre> tag
  $pageText =~
    s/\&lt;pre\&gt;((.|\n)*?)\&lt;\/pre\&gt;/&StorePre($1)/ige;
  $pageText =~ s/\\ *\r?\n/ /g;       # Join lines separated by backslash
  # Use images, non-line-oriented markup only
  $pageText = &CommonMarkup($pageText, 1, 0);
  foreach (split(/\n/, $pageText)) {  # Process lines one-at-a-time
    $_ .= "\n";
    # Start-of-line markup patterns
    $SaveHtmlStack = 0;
    s/^\s*$/<p>\n/                   && ($SaveHtmlStack = 1);
    s/^(\;+)([^:]+\:?)\:/<dt>$2<dd>/ && &StackHtml("DL", length $1);
    s/^(\:+)/<dt><dd>/               && &StackHtml("DL", length $1);
    s/^(\*+)/<li>/                   && &StackHtml("UL", length $1);
    s/^(\#+)/<li>/                   && &StackHtml("OL", length $1);
    /^\s/                            && &StackHtml("PRE", 1);
    &UnstackHtml() if (!$SaveHtmlStack);
    # Do only line-oriented markup (image flag is unused)
    $PageHtml .= &CommonMarkup($_, 1, 2);
  }
  &UnstackHtml();
  $PageHtml =~ s/$FS(\d+)$FS/$SaveUrl{$1}/ge;
  return $PageHtml;
}

sub StoreRaw {
  my ($html) = @_;

  $SaveUrl{$SaveUrlIndex} = $html;
  return $FS . $SaveUrlIndex++ . $FS;
}

sub StorePre {
  my ($html) = @_;

  return &StoreRaw("<pre>" . $html . "</pre>");
}

sub StoreHref {
  my ($anchor, $text) = @_;

  return "<a" . &StoreRaw($anchor) . ">$text</a>";
}

sub StoreUrl {
  my ($name, $useImage) = @_;
  my ($link, $extra);

  ($link, $extra) = &UrlLink($name, $useImage);
  # Next line ensures no empty links are stored
  $link = &StoreRaw($link)  if ($link ne "");
  return $link . $extra;
}

sub UrlLink {
  my ($rawname, $useImage) = @_;
  my ($name, $punct);

  ($name, $punct) = &SplitUrlPunct($rawname);
  if ($useImage && ($name =~ /\.$ImageExtensions$/)) {
    return ("<img src=\"$name\">", $punct);
  }
  return ("<a href=\"$name\">$name</a>", $punct);
}

sub StoreBracketUrl {
  my ($name) = @_;
  my ($index);

  $index = &GetBracketUrlIndex($name);
  return &StoreRaw("<a href=\"$name\">[$index]</a>");
}

sub StoreRFC {
  my ($num) = @_;

  return &StoreRaw(&RFCLink($num));
}

sub RFCLink {
  my ($num) = @_;

  return "<a href=\"http://www.faqs.org/rfcs/rfc${num}.html\">RFC $num</a>";
}

sub StoreISBN {
  my ($num) = @_;

  return &StoreRaw(&ISBNLink($num));
}

sub ISBNLink {
  my ($rawnum) = @_;
  my ($rawprint, $html, $num, $first, $second, $third); 

  $num = $rawnum;
  $rawprint = $rawnum;
  $rawprint =~ s/ +$//;
  $num =~ s/[- ]//g;
  if (length($num) != 10) {
    return "ISBN $rawnum";
  }

  $first  = "<a href=\"http://shop.barnesandnoble.com/bookSearch/"
            . "isbnInquiry.asp?isbn=$num\">";
  $second = "<a href=\"http://www.amazon.com/exec/obidos/"
            . "ISBN=$num\">Amazon</a>";
  $third  = "<a href=\"http://www.pricescan.com/books/"
            . "BookDetail.asp?isbn=$num\">Pricescan</a>";

  $html  = $first . "ISBN " . $rawprint . "</a> ";
  $html .= "($second, $third)";
  $html .= " "  if ($rawnum =~ / $/);  # Add space if old ISBN had space.
  return $html;
}

sub SplitUrlPunct {
  my ($url) = @_;
  my ($punct);

  if ($url =~ s/\"\"$//) {
    return ($url, "");
  }
  $punct = "";
  ($punct) = ($url =~ /([^a-zA-Z0-9\/\xc0-\xff]+)$/);
  $url =~ s/([^a-zA-Z0-9\/\xc0-\xff]+)$//;
  return ($url, $punct);
}

sub StripUrlPunct {
  my ($url) = @_;
  my ($junk);

  ($url, $junk) = &SplitUrlPunct($url);
  return $url;
}

sub ReportError {
  my ($errmsg) = @_;

  print $q->header, "<H2>", $errmsg, "</H2>", $q->end_html;
}

sub ValidId {
  my ($id) = @_;

  if (!($id =~ /^$LinkPattern$/)) {
    ReportError("Invalid Page $id");
    return 0;
  }
  if (length($id) > 120) {
    ReportError("Page name is too long: $id");
    return 0;
  }
  return 1;
}

sub OpenNewPage {
  my ($id) = @_;
  my $data;

  %page = ();
  $page{'text'} = "Describe the new page here.\n";
  $page{'timestamp'} = 943410730; # Must be fixed for edit conflicts.
  $page{'revision'} = 0;
  $page{'majorcopy'} = "";
  $page{'authorcopy'} = "";
  $page{'minordiff'} = "";
  $page{'majordiff'} = "1";
  $page{'authordiff'} = "1";
  $FullPage{'current'} = join($FS3, %page);
  $FullPage{dbrev} = 2;
}

sub OpenPage {
  my ($id) = @_;
  my ($fname, $data);

  if ($OpenId eq $id) {
    return;
  }
  $fname = &GetPageDirectory($id) . "$id.db";
  if (-f $fname) {
    open(IN, "<$fname") or die "Can't open $fname: $!";
    $data=<IN>;
    close IN;
    %FullPage = split(/$FS1/, $data, -1);  # -1 keeps trailing null fields
  } else {
    &OpenNewPage($id);
  }
  if ($FullPage{dbrev} != 2) {
    &UpdatePageVersion();
  }
  $OpenId = $id;
}

sub UpdatePageVersion {
  my ($oldversion, %pg);

  $oldversion = $FullPage{dbrev};
  %pg = split(/$FS3/, $FullPage{current}, -1);
  if ($oldversion < 2) {   # Initial version "1.1"
    # The upgraded page is considered to be a newly-created page.
    $pg{'majorcopy'} = $pg{'text'};
    $pg{'authorcopy'} = $pg{'text'};
    $pg{'minordiff'} = $pg{'prevdiff'};
    $pg{'majordiff'} = "1";
    $pg{'authordiff'} = "1";
    undef $pg{'prevdiff'};
    undef $FullPage{'authorcopy'};
    $FullPage{'current'} = join($FS3, %pg);
    $FullPage{dbrev} = 2;
  }
}

# Use this to get single-version pages.
sub GetPage {
  my ($id, $version) = @_;

  &OpenPage($id);
  return split(/$FS3/, $FullPage{$version}, -1);
}

sub GetCurrentPage {
  my ($id) = @_;

  return &GetPage($id,'current');
}

# Failing to get a cache lock is not a serious problem.
sub RequestCacheLock {
  my $n = 0;
  while (mkdir($LockDir, 0555) == 0) {
    $! == 17 || return 0; # EEXIST == 17 is OK, try later.
    $n++ < 4 || return 0;
    sleep(2);
  }
  return 1;
}

sub RequestLock {
  my $n = 0;
  while (mkdir($LockDir, 0555) == 0) {
    # EEXIST == 17 is OK, try later.
    $! ==  17 || die("can't make $LockDir: $!\n");
    $n++ < 10 || die("timed out waiting for $LockDir\n");
    sleep(3);
  }
}

sub ReleaseLock {
  rmdir($LockDir);
}

sub WriteStringToFile {
  my ($file, $string) = @_;

  open (OUT, ">$file") or die ("cant write $file: $!");
  print OUT  $string;
  close(OUT);
}

sub CreatePageDir {
  my ($dir, $id) = @_;
  my $subdir;

  $subdir = $dir . "/" . &GetPageDirectory($id);
  mkdir($subdir,0775)  if (!(-d $subdir));
  if ($id =~ m|([^/]+)/|) {
    $subdir = $subdir . "/" . $1;
    mkdir($subdir,0775)  if (!(-d $subdir));
  }
}

sub UpdateHtmlCache {
  my ($id, $html) = @_;
  my $id_file;

  $id_file = $HtmlDir . "/" . &GetPageDirectory($id) . "$id.htm";
  mkdir($HtmlDir,0775)  if (!(-d $HtmlDir));
  &CreatePageDir($HtmlDir, $id);
  if (&RequestCacheLock()) {
    &WriteStringToFile($id_file, $html);
    &ReleaseLock();
  }
}

sub CalcDay {
  my ($ts) = @_;

  $ts += $TimeZoneOffset;
  my ($sec, $min, $hour, $mday, $mon, $year) = localtime($ts);

  return ("January", "February", "March", "April", "May", "June",
          "July", "August", "September", "October", "November",
          "December")[$mon]. " " . $mday . ", " . ($year+1900);
}

sub CalcDayNow {
  return CalcDay($^T);
}

sub CalcTime {
  my ($ts) = @_;

  $ts += $TimeZoneOffset;
  my ($sec, $min, $hour, $mday, $mon, $year) = localtime($ts);
  my ($ampm, $mytz);

  $mytz = "";
  if (($TimeZoneOffset == 0) && ($ScriptTZ ne "")) {
    $mytz = " " . $ScriptTZ;
  }
  $ampm = "am";
  if ($hour>11) {
    $ampm = "pm";
    $hour = $hour - 12;
  }
  $hour = 12   if ($hour == 0);
  $min = "0" . $min   if ($min<10);
  return $hour . ":" . $min . " " . $ampm . $mytz;
}

sub TimeToText {
  my ($t) = @_;

  return &CalcDay($t) . " " . &CalcTime($t);
}

sub ActionRc {
  my ($filedata, $rcline, $i, $daysago, $lastts, $ts, $idOnly);
  my @fullrc = ();
  my $starttime = 0;
  my $showbar = 0;

  if (!open(IN, "<$RcFile")) {
    print "<p><strong>Could not open RecentChanges log file: </strong>";
    print "$RcFile<p>Error was:\n<pre>$!</pre>\n";
    print "<p>Note: This error is normal if no changes have been made.\n";
    return;
  }
  $filedata=<IN>;
  close IN;
  @fullrc = split(/\n/, $filedata);
  $lastts = 0;
  if (@fullrc > 0) {  # Only false if no lines in file
    ($lastts) = split(/$FS3/, $fullrc[$#fullrc]);
  }
  $lastts++  if (($^T - $lastts) > 5);  # Skip last unless very recent

  if (&GetParam("from", 0)) {
    $starttime = &GetParam("from", 0);
    print "<h2>Updates since ", &TimeToText($starttime), "</h2>\n";
  } else {
    $daysago = &GetParam("days", 0);
    $daysago = &GetParam("rcdays", 0)  if ($daysago == 0);
    if ($daysago) {
      $starttime = $^T - ((24*60*60)*$daysago);
      print "<h2>Updates in the last $daysago day",
            (($daysago != 1)?"s":""), "</h2>\n";
    }
  }
  if ($starttime == 0) {
    $starttime = $^T - ((24*60*60)*$RcDefault);
    print "<h2>Updates in the last $RcDefault days</h2>\n";
  }
  $idOnly = &GetParam("rcidonly", "");
  if ($idOnly ne "") {
    print "<b>(for " . &ScriptLink($idOnly, $idOnly)
          . " only)</b><br>";
  }

  foreach $i (@RcDays) {
    print " | "  if $showbar;
    $showbar = 1;
    print &ScriptLink("action=rc&days=$i", "$i day" . (($i != 1)?"s":""));
  }
  print "<br>" . &ScriptLink("action=rc&from=$lastts",
                             "List new changes starting from");
  print " " . &TimeToText($lastts), "<br>\n";

  # Later consider a binary search?
  $i = 0;
  while ($i < @fullrc) {  # Optimization: skip old entries quickly
    ($ts) = split(/$FS3/, $fullrc[$i]);
    if ($ts >= $starttime) {
      $i -= 1000  if ($i > 0);
      last;
    }
    $i += 1000;
  }
  $i -= 1000  if (($i > 0) && ($i >= @fullrc));
  for (; $i < @fullrc ; $i++) {
    ($ts) = split(/$FS3/, $fullrc[$i]);
    last if ($ts >= $starttime);
  }
  if ($i == @fullrc) {
    print "<br><strong>No updates since ", &TimeToText($starttime),
          "</strong><br>\n";
  } else {
    splice(@fullrc, 0, $i);  # Remove items before index $i
    # Later consider an end-time limit (items older than X)
    print &GetRcHtml(@fullrc);
  }
  print "<p>Page generated ", &TimeToText($^T), "<br>\n";
}

sub GetRcHtml {
  my @outrc = @_;
  my ($rcline, $html, $date, $sum, $edit, $count, $newtop, $author);
  my ($showedit, $inlist, $link, $all, $idOnly);
  my ($ts, $pagename, $summary, $isEdit, $host, $kind, $extraTemp);
  my %extra = ();
  my %changetime = ();
  my %pagecount = ();

  $showedit = &GetParam("rcshowedit", $ShowEdits);
  $showedit = &GetParam("showedit", $showedit);
  if ($showedit != 1) {
    my @temprc = ();
    foreach $rcline (@outrc) {
      ($ts, $pagename, $summary, $isEdit, $host) = split(/$FS3/, $rcline);
      if ($showedit == 0) {  # 0 = No edits
        push(@temprc, $rcline)  if (!$isEdit);
      } else {               # 2 = Only edits
        push(@temprc, $rcline)  if ($isEdit);
      }
    }
    @outrc = @temprc;
  }

  # Later consider folding into loop above?
  # Later add lines to assoc. pagename array (for new RC display)
  foreach $rcline (@outrc) {
    ($ts, $pagename) = split(/$FS3/, $rcline);
    $pagecount{$pagename}++;
    $changetime{$pagename} = $ts;
  }
  $date = "";
  $inlist = 0;
  $html = "";
  $all = &GetParam("rcall", 0);
  $all = &GetParam("all", $all);
  $newtop = &GetParam("rcnewtop", $RecentTop);
  $newtop = &GetParam("newtop", $newtop);
  $idOnly = &GetParam("rcidonly", "");

  @outrc = reverse @outrc if ($newtop);
  foreach $rcline (@outrc) {
    ($ts, $pagename, $summary, $isEdit, $host, $kind, $extraTemp)
      = split(/$FS3/, $rcline);
    # Later: need to change $all for new-RC?
    next  if ((!$all) && ($ts < $changetime{$pagename}));
    next  if (($idOnly ne "") && ($idOnly ne $pagename));
    %extra = split(/$FS2/, $extraTemp, -1);
    if ($date ne &CalcDay($ts)) {
      $date = &CalcDay($ts);
      if ($inlist) {
        $html .= "</UL>\n";
        $inlist = 0;
      }
      $html .= "<p><strong>" . $date . "</strong><p>\n";
    }
    if (!$inlist) {
      $html .= "<UL>\n";
      $inlist = 1;
    }
    $host = &QuoteHtml($host);
    if (defined($extra{'name'}) && defined($extra{'id'})) {
      $author = &GetAuthorLink($host, $extra{'name'}, $extra{'id'});
    } else {
      $author = &GetAuthorLink($host, "", 0);
    }
    $sum = "";
    if (($summary ne "") && ($summary ne "*")) {
      $summary = &QuoteHtml($summary);
      $sum = "<strong>[$summary]</strong> ";
    }
    $edit = "";
    $edit = "<em>(edit)</em> "  if ($isEdit);
    $count = "";
    if ((!$all) && ($pagecount{$pagename} > 1)) {
      $count = "($pagecount{$pagename} changes) ";
    }
    $link = "";
    if ($UseDiff && &GetParam("diffrclink", 1)) {
      $link .= &ScriptLinkDiff(4, $pagename, "(diff)") . "  ";
    }
    $link .= &ScriptLink($pagename, $pagename);
    $html .= "<li>$link ";
    # Later do new-RC looping here.
    $html .=  &CalcTime($ts) . " $count$edit" . " $sum";
    $html .= "$author\n";
  }
  $html .= "</UL>\n" if ($inlist);
  return $html;
}

sub GetAuthorLink {
  my ($host, $userName, $uid) = @_;
  my ($html, $title);

  if (!($userName =~ /^$LinkPattern$/)) {  # Invalid under current rules
    $userName = "";  # Just pretend it isn't there.
  }
  $html  = ". . . . . ";  # Later optional
  # Later have user preference for link titles and/or host text?
  if (($uid > 0) && ($userName ne "")) {
    $html .= &ScriptLinkTitle($userName, $userName, "ID $uid from $host");
  } else {
    $html .= $host;
  }
  return $html;
}

sub GetRedirectPage {
  my ($newid) = @_;
  my ($url, $html);

  # Normally get URL from script, but allow override.
  $FullUrl = $q->url(-full=>1)  if ($FullUrl eq "");
  $url = $FullUrl . "?" . $newid;
  $html = $q->redirect(-uri=>$url, -method=>"GET");
  $html .= "\nYour browser should go to the $newid page.";
  $html .= "  If it does not, click <a href=\"$url\">$newid</a>";
  $html .= " to continue.";
  return $html;
}
#END_OF_BROWSE_CODE

# == Page-editing and other special-action code ========================
$OtherCode = ""; # Comment next line to always compile (slower)
$OtherCode = <<'#END_OF_OTHER_CODE';
use Socket;  # For host lookups

sub DoOtherRequest {
  my ($id, $action, $text, $search);

  $action = &GetParam("action", "");
  $id = &GetParam("id", "");
  if ($action ne "") {
    $action = lc($action);
    if      ($action eq "edit") {
      &ActionEdit($id, 0, 0, 0, "", 0)  if &ValidId($id);
    } elsif ($action eq "editcopy") {
      &ActionEdit($id, 1, 0, 0, "", 0)  if &ValidId($id);
    } elsif ($action eq "unlock") {
      &DoUnlock();
    } elsif ($action eq "index") {
      &DoIndex();
    } elsif ($action eq "random") {
      &DoRandom();
    } elsif ($action eq "links") {
      &DoLinks();
    } elsif ($action eq "editprefs") {
      &DoEditPrefs();
#   } elsif ($action eq "convert") {  # Normally not available
#     &DoConvert();
    } elsif ($action eq "login") {
      &DoEnterLogin();
    } elsif ($action eq "newlogin") {
      $UserID = 0;
      &DoEditPrefs();  # Also creates new ID
    } else {
      &ReportError("Invalid action parameter $action");
    }
    return;
  }
  if (&GetParam("edit_prefs", 0)) {
    &DoUpdatePrefs();
    return;
  }
  if (&GetParam("enter_login", 0)) {
    &DoLogin();
    return;
  }
  $search = &GetParam("search", "");
  if (($search ne "") || (&GetParam("dosearch", "") ne "")) {
    &DoSearch($search);
    return;
  }
  # Handle posted pages
  if (&GetParam("oldtime", "") ne "") {
    $id = &GetParam("title", "");
    &ActionPost()  if &ValidId($id);
    return;
  }
  &ReportError("Invalid URL.");
}

sub ActionEdit {
  my ($id, $editCopy, $isConflict, $oldTime, $newText, $preview) = @_;
  my $oldText = "";
  my %pg = ();
  my ($header, $editRows, $editCols, $userName);

  if (!$EditAllowed) {
    &ReportError("Editing not allowed: $SiteName is read-only.");
    return;
  }
  %pg = &GetCurrentPage($id);
  $header = "Editing $id";
  if ($editCopy) {
    $oldText = $pg{'authorcopy'};
    $header = "Editing Old Copy of $id";
  } else {
    $oldText = $pg{'text'};
  }
  if ($preview && !$isConflict) {
    $oldText = $newText;
  }
  $editRows = &GetParam("editrows", 20);
  $editCols = &GetParam("editcols", 65);
  print &GetHeader("", &QuoteHtml($header), "");
  if ($isConflict) {
    $editRows -= 10  if ($editRows > 19);
    print "\n<H1>Edit Conflict!</H1>\n";
    if ($isConflict>1) {
      # The main purpose of a new warning is to display more text
      # and move the save button down from its old location.
      print "\n<H2>(This is a new conflict)</H2>\n";
    }
    print "<p><strong>Someone saved this page after you started editing.",
          " The top textbox contains the saved text.",
          " Only the text in the top textbox will be saved.</strong><br>\n",
          " Scroll down to see your edited version.<br>\n",
          "Last save time: ", &TimeToText($oldTime),
          " (Current time is: ", &TimeToText($^T), ")<br>\n";
  }
  print &GetFormStart();
  print &GetHiddenValue("title", $id), "\n",
        &GetHiddenValue("oldtime", $pg{'timestamp'}), "\n",
        &GetHiddenValue("oldconflict", $isConflict), "\n";
  print &GetTextArea('text', $oldText, $editRows, $editCols);
  print "<p>Summary:",
        $q->textfield(-name=>'summary',
                      -default=>'*', -override=>1,
                      -size=>60, -maxlength=>200);
  print "<br>", $q->checkbox(-name=>'recent_edit',
                             -label=>'This change is a minor edit.');
  if ($UseEmailNotify) {
    print "&nbsp;&nbsp;&nbsp;" .
           $q->checkbox(-name=> 'do_email_notify',
             -label=>"Send email notification that $id has been changed." );
  }
  print "<br>", $q->submit(-name=>'Save', -value=>'Save'), "\n";
  $userName = &GetParam("username", "");
  if ($userName ne "") {
    print " (Your user name is " . &GetPageLink($userName) . ") ";
  } else {
    print " (Visit " . &GetPrefsLink() . " to set your user name.) ";
  }
  print $q->submit(-name=>'Preview', -value=>'Preview'), "\n";

  if ($isConflict) {
    print "\n<br><hr><p><strong>This is the text you submitted:",
          "</strong><p>",
          &GetTextArea('newtext', $newText, $editRows, $editCols),
          "<p>\n";
  }
  print "<hr>\n";
  if ($preview) {
    print "<h2>Preview:</h2>\n";
    if ($isConflict) {
      print "<b>NOTE: This preview shows the other author's version.",
            "</b><hr>\n";
    }
    $MainPage = $id;
    $MainPage =~ s|/.*||;  # Only the main page name (remove subpage)
    print &WikiToHTML($oldText) . "<hr>\n";
  }
  if ((!$editCopy)
       && defined($pg{'authorcopy'})
       && ($pg{'authorcopy'} ne "")) {
    print &ScriptLink("action=editcopy&id=$id",
                      "Edit copy from previous author") . "<br>\n";
  }
  print &GetGotoBar($id);
  print $q->endform;
  print $q->end_html;
}

sub GetTextArea {
  my ($name, $text, $rows, $cols) = @_;

  if (&GetParam("editwide", 1)) {
    return $q->textarea(-name=>$name, -default=>$text,
                        -rows=>$rows, -columns=>$cols, -override=>1,
                        -style=>'width:100%', -wrap=>'virtual');
  }
  return $q->textarea(-name=>$name, -default=>$text,
                      -rows=>$rows, -columns=>$cols, -override=>1,
                      -wrap=>'virtual');
}

sub DoEditPrefs {
  my ($check, %labels);

  &DoNewLogin()  if ($UserID < 400);
  print &GetHeader("", "Editing Preferences", "");
  print &GetFormStart();
  print GetHiddenValue("edit_prefs", 1), "\n";
  print "<b>User Information:</b>\n";
  print "<br>Your User ID number: $UserID\n";
  print "<br>UserName: ", &GetFormText('username', "", 20, 50);
  print " (blank to remove, or valid page name)";
  print "<br>Set Password: ",
        $q->password_field(-name=>'p_password', -value=>'*', 
                           -size=>15, -maxlength=>50),
        " (blank to remove password)",
        "<br>(Passwords are only used for sharing user IDs",
        " and preferences between multiple systems.",
        " Passwords are completely optional.)";
  if ($UseEmailNotify) {
    print "<br>";
    print &GetFormCheck('notify', 1,
      'Include this address in the site email list. '
      . '(Uncheck the box to remove the address.)');
    print "<br> Email Address: ", &GetFormText('email', "", 30, 60);
  }
  print "<hr><b>RecentChanges:</b>\n";
  print "<br>Default days to display: ",
        &GetFormText('rcdays', $RcDefault, 4, 9);
  print "<br>", &GetFormCheck('rcnewtop', $RecentTop,
                              'Most recent changes on top');
  print "<br>", &GetFormCheck('rcall', 0,
                              'Show all changes (not just most recent)');
  %labels = (0=>'Hide minor edits', 1=>'Show minor edits',
             2=>'Show only minor edits');
  print "<br>Minor edit display: ";
  print $q->popup_menu(-name=>'p_rcshowedit',
                       -values=>[0,1,2], -labels=>\%labels,
                       -default=>&GetParam("rcshowedit", $ShowEdits));
  if ($UseDiff) {
    print "<hr><b>Differences:</b>\n";
    print "<br>", &GetFormCheck('diffrclink', 1,
                                'Show (diff) links on RecentChanges');
    print "<br>", &GetFormCheck('alldiff', 0,
                                'Show differences on all pages');
    print "  (",  &GetFormCheck('norcdiff', 1,
                                'No differences on RecentChanges'), ")";
    %labels = (1=>'Major', 2=>'Minor', 3=>'Author');
    print "<br>Default difference type: ";
    print $q->popup_menu(-name=>'p_defaultdiff',
                         -values=>[1,2,3], -labels=>\%labels,
                         -default=>&GetParam("defaultdiff", 1));
  }
  print "<hr><b>Misc:</b>\n";
  print "<br>Server time: ", &TimeToText($^T-$TimeZoneOffset);
  print "<br>Time Zone offset (hours): ",
        &GetFormText('tzoffset', 0, 4, 9);
  print "<br>", &GetFormCheck('editwide', 1,
                              'Use 100% wide edit area (if supported)');
  print "<br>Edit area rows: ", &GetFormText('editrows', 20, 4, 4),
        " columns: ", &GetFormText('editcols', 65, 4, 4);

  print "<br>", &GetFormCheck('toplinkbar', 1,
                              'Show link bar on top');
  print "<br>", &GetFormCheck('linkrandom', 0,
                              'Add "Random Page" link to link bar');
  print "<br>", $q->submit(-name=>'Save'), "\n";
  print "<hr>\n";
  print &GetGotoBar("");
  print $q->endform;
# print "<br>cookie id: |" . $UserCookie{'id'} . "|<br>" .  # debug
#       "raw cookie: |" . $ENV{'HTTP_COOKIE'} . "|<br>";
  print $q->end_html;
}

sub GetFormText {
  my ($name, $default, $size, $max) = @_;
  my $text = &GetParam($name, $default);

  return $q->textfield(-name=>"p_$name", -default=>$text,
                       -override=>1, -size=>$size, -maxlength=>$max);
}

sub GetFormCheck {
  my ($name, $default, $label) = @_;
  my $checked = (&GetParam($name, $default) > 0);

  return $q->checkbox(-name=>"p_$name", -override=>1, -checked=>$checked,
                      -label=>$label);
}

sub DoUpdatePrefs {
  my ($username, $password);

  # All link bar settings should be updated before printing the header
  &UpdatePrefCheckbox("toplinkbar");
  &UpdatePrefCheckbox("linkrandom");
  print &GetHeader("","Saving Preferences", "");
  print "<br>";
  if ($UserID < 1001) {
    print "<b>Invalid UserID $UserID, preferences not saved.</b>";
    if ($UserID == 111) {
      print "<br>(Preferences require cookies, but no cookie was sent.)";
    }
    &PrintCommonFooter();
    return;
  }
  $username = &GetParam("p_username",  "");
  if ($username eq "") {
    print "UserName removed.<br>";
    undef $UserData{'username'};
  } elsif (!($username =~ /^$LinkPattern$/)) {  # Currently invalid 
    print "Invalid UserName $username: not saved.<br>\n";
  } elsif (length($username) > 50) {  # Too long
    print "UserName must be 50 characters or less. (not saved)<br>\n";
  } else {
    print "UserName $username saved.<br>";
    $UserData{'username'} = $username;
  }
  $password = &GetParam("p_password",  "");
  if ($password eq "") {
    print "Password removed.<br>";
    undef $UserData{'password'};
  } elsif ($password ne "*") {
    print "Password changed.<br>";
    $UserData{'password'} = $password;
  }
  if ($UseEmailNotify) {
    &UpdatePrefCheckbox("notify");
    &UpdateEmailList();
  }
  &UpdatePrefNumber("rcdays", 0, 0, 999999);
  &UpdatePrefCheckbox("rcnewtop");
  &UpdatePrefCheckbox("rcall");
  &UpdatePrefCheckbox("editwide");
  if ($UseDiff) {
    &UpdatePrefCheckbox("norcdiff");
    &UpdatePrefCheckbox("diffrclink");
    &UpdatePrefCheckbox("alldiff");
    &UpdatePrefNumber("defaultdiff", 1, 1, 3);
  }
  &UpdatePrefNumber("rcshowedit", 1, 0, 2);
  &UpdatePrefNumber("tzoffset", 0, -999, 999);
  &UpdatePrefNumber("editrows", 1, 1, 999);
  &UpdatePrefNumber("editcols", 1, 1, 999);
  print "Server time: ", &TimeToText($^T-$TimeZoneOffset), "<br>";
  $TimeZoneOffset = &GetParam("tzoffset", 0) * (60 * 60);
  print "Local time: ", &TimeToText($^T), "<br>";

  &SaveUserData();
  print "<b>Preferences saved.</b>";
  &PrintCommonFooter();
}

# add or remove email address from preferences to $DatDir/emails
sub UpdateEmailList {
  my (@old_emails);

  local $/ = "\n";  # don't slurp whole files in this sub.
  if (my $new_email = $UserData{'email'} = &GetParam("p_email", "")) {
    my $notify = $UserData{'notify'};
    if (-f "$DataDir/emails") {
      open(NOTIFY, "$DataDir/emails")
        or die "Couldn't read from $DataDir/email: $!\n";
      @old_emails = <NOTIFY>;
      close(NOTIFY);
    } else {
      @old_emails = ();
    }
    my $already_in_list = grep /$new_email/, @old_emails;
    if ($notify and (not $already_in_list)) {
      &RequestLock();
      open(NOTIFY, ">>$DataDir/emails")
        or die "Couldn't append to $DataDir/email: $!\n";
      print NOTIFY $new_email, "\n";
      close(NOTIFY);
      &ReleaseLock();
    }
    elsif ((not $notify) and $already_in_list) {
      &RequestLock();
      open(NOTIFY, ">$DataDir/emails")
        or die "Couldn't overwrite $DataDir/email: $!\n";
      foreach (@old_emails) {
        print NOTIFY "$_" unless /$new_email/;
      }
      close(NOTIFY);
      &ReleaseLock();
    }
  }
}

sub UpdatePrefCheckbox {
  my ($param) = @_;
  my $temp = &GetParam("p_$param", "*");

  $UserData{$param} = 1  if ($temp eq "on");
  $UserData{$param} = 0  if ($temp eq "*");
  # It is possible to skip updating by using another value, like "2"
}

sub UpdatePrefNumber {
  my ($param, $integer, $min, $max) = @_;
  my $temp = &GetParam("p_$param", "*");

  return  if ($temp eq "*");
  $temp =~ s/[^-\d\.]//g;
  $temp =~ s/\..*//  if ($integer);
  return  if ($temp eq "");
  return  if (($temp < $min) || ($temp > $max));
  $UserData{$param} = $temp;
  # Later consider returning status?
}

sub PrintCommonFooter {
  print "<hr>", &GetFormStart(),
        &GetGotoBar(""),
        &GetSearchForm(),
        $q->endform, $q->end_html;
}

sub DoIndex {
  print &GetHeader("","Index of all pages", "");
  print "<br>";
  &PrintPageList(&AllPagesList());
  &PrintCommonFooter();
}

sub DoRandom {
  my ($id, @pageList);

  @pageList = &AllPagesList();  # Optimize?
  $id = $pageList[int(rand($#pageList + 1))];
  &ReBrowsePage($id, "");
}

# Convert old tab-based formatting to new formats.
sub DoConvert {
  my ($name, @found, %pg);

  &RequestLock();
  print &GetHeader("","Converting all pages", "");
  print "<br>";
  @found = ();
  foreach $name (&AllPagesList()) {
    %pg = &GetCurrentPage($name);
    push(@found, $name);
    $pg{text} =~ s/(^|\n)(\t+)\s+:\t/$1 . (":" x (length $2))/ge;
    $pg{text} =~
      s/(^|\n)(\t+)(.+):\t/$1 . (";" x (length $2)) . $3 . ":"/ge;
    $pg{text} =~ s/(^|\n)(\t+)\*/$1 . ("*" x (length $2))/ge;
    $pg{text} =~ s/(^|\n)(\t+)\d+\.?/$1 . ("#" x (length $2))/ge;
    $FullPage{'current'} = join($FS3, %pg);
    &SavePage($name);
  }
  &PrintPageList(@found);
  &PrintCommonFooter();
  &ReleaseLock();
}

# Create a new user file/cookie pair
sub DoNewLogin {
  # Later consider warning if cookie already exists
  # (maybe use "replace=1" parameter)
  &CreateUserDir();
  $SetCookie{'id'} = &GetNewUserId;
  $SetCookie{'randkey'} = int(rand(1000000000));
  $SetCookie{'rev'} = 1;
  %UserCookie = %SetCookie;
  $UserID = $SetCookie{'id'};
  # The cookie will be transmitted in the next header
  %UserData = %UserCookie;
  $UserData{'createtime'} = $^T;
  $UserData{'createip'} = $ENV{REMOTE_ADDR};
  &SaveUserData();
}

sub DoEnterLogin {
  print &GetHeader("", "Login", "");
  print &GetFormStart();
  print GetHiddenValue("enter_login", 1), "\n";
  print "<br>User ID number: ",
        $q->textfield(-name=>'p_userid', -value=>'',
                      -size=>15, -maxlength=>50);
  print "<br>Password: ",
        $q->password_field(-name=>'p_password', -value=>'', 
                           -size=>15, -maxlength=>50);
  print "<br>", $q->submit(-name=>'Login'), "\n";
  print "<hr>\n";
  print &GetGotoBar("");
  print $q->endform;
  print $q->end_html;
}

sub DoLogin {
  my ($uid, $password, $success);

  $success = 0;
  $uid = &GetParam("p_userid", "");
  $uid =~ s/\D//g;
  $password = &GetParam("p_password",  "");
  if (($uid > 199) && ($password ne "") && ($password ne "*")) {
    $UserID = $uid;
    &LoadUserData();
    if ($UserID > 199) {
      if (defined($UserData{'password'}) &&
          ($UserData{'password'} eq $password)) {
        $SetCookie{'id'} = $uid;
        $SetCookie{'randkey'} = $UserData{'randkey'};
        $SetCookie{'rev'} = 1;
        $success = 1;
      }
    }
  }
  print &GetHeader("", "Login Results", "");
  if ($success) {
    print "Login for user ID $uid complete.";
  } else {
    print "Login for user ID $uid failed.";
  }
  print "<hr>\n";
  print &GetGotoBar("");
  print $q->endform;
  print $q->end_html;
}

sub GetNewUserId {
  my ($id);

  $id = 1001;
  while (-f &UserDataFilename($id+1000)) {
    $id += 1000;
  }
  while (-f &UserDataFilename($id+100)) {
    $id += 100;
  }
  while (-f &UserDataFilename($id+10)) {
    $id += 10;
  }
  &RequestLock();
  while (-f &UserDataFilename($id)) {
    $id++;
  }
  &WriteStringToFile(&UserDataFilename($id), "lock");  # reserve the ID
  &ReleaseLock();
  return $id;
}

# Later get user-level lock
sub SaveUserData {
  my ($userFile, $data);

  &CreateUserDir();
  $userFile = &UserDataFilename($UserID);
  $data = join($FS1, %UserData);
  &WriteStringToFile($userFile, $data);
}

sub CreateUserDir {
  my ($n, $subdir);

  if (!(-d "$UserDir/0")) {
    mkdir($UserDir,0775)  if (!(-d $UserDir));

    foreach $n (0..9) {
      $subdir = "$UserDir/$n";
      mkdir($subdir,0775)  if (!(-d $subdir));
    }
  }
}

sub DoSearch {
  my ($string) = @_;

  if ($string eq "") {
    &DoIndex();
    return;
  }
  print &GetHeader("",&QuoteHtml("Search for: $string"), "");
  print "<br>";
  &PrintPageList(&SearchTitleAndBody($string));
  &PrintCommonFooter();
}

sub PrintPageList {
  my $pagename;

  print "<h2>", ($#_ + 1), " pages found:</h2>\n";
  foreach $pagename (@_) {
    print ".... "  if ($pagename =~ m|/|);
    print &GetPageLink($pagename), "<br>\n";
  }
}

sub DoLinks {
  print &GetHeader("",&QuoteHtml("Full Link List"), "");
  print "<hr><pre>\n\n\n\n\n";  # Extra lines to get below the logo
  &PrintLinkList(&GetFullLinkList());
  print "</pre>\n";
  print $q->end_html;
}

sub PrintLinkList {
  my ($pagelines, $page, $names, $editlink);
  my ($link, $extra, @links, %pgExists);

  %pgExists = ();
  foreach $page (&AllPagesList()) {
    $pgExists{$page} = 1;
  }
  $names = &GetParam("names", 1);
  $editlink = &GetParam("editlink", 0);
  foreach $pagelines (@_) {
    @links = ();
    foreach $page (split(' ', $pagelines)) {
      if ($page =~ /\:/) {  # URL or InterWiki form
        if ($page =~ /$UrlPattern/) {
          ($link, $extra) = &UrlLink($page);
        } else {
          ($link, $extra) = &InterPageLink($page);
        }
      } else {
        if ($pgExists{$page}) {
          $link = &GetPageLink($page);
        } else {
          $link = $page;
          if ($editlink) {
            $link .= &GetEditLink($page, "?");
          }
        }
      }
      push(@links, $link);
    }
    if (!$names) {
      shift(@links);
    }
    print join(' ', @links), "\n";
  }
}

sub GetFullLinkList {
  my ($name, $unique, $sort, $exists, $empty, $link, $search);
  my ($pagelink, $interlink, $urllink);
  my (@found, @links, @newlinks, @pglist, %pg, %pgExists, %seen);

  $unique = &GetParam("unique", 1);
  $sort = &GetParam("sort", 1);
  $pagelink = &GetParam("page", 1);
  $interlink = &GetParam("inter", 0);
  $urllink = &GetParam("url", 0);
  $exists = &GetParam("exists", 2);
  $empty = &GetParam("empty", 0);
  $search = &GetParam("search", "");
  if (($interlink == 2) || ($urllink == 2)) {
    $pagelink = 0;
  }

  %pgExists = ();
  @pglist = &AllPagesList();
  foreach $name (@pglist) {
    $pgExists{$name} = 1;
  }
  %seen = ();
  foreach $name (@pglist) {
    @newlinks = ();
    if ($unique != 2) {
      %seen = ();
    }
    @links = &GetPageLinks($name, $pagelink, $interlink, $urllink);

    foreach $link (@links) {
      $seen{$link}++;
      if (($unique > 0) && ($seen{$link} != 1)) {
        next;
      }
      if (($exists == 0) && ($pgExists{$link} == 1)) {
        next;
      }
      if (($exists == 1) && ($pgExists{$link} != 1)) {
        next;
      }
      if (($search ne "") && !($link =~ /$search/)) {
        next;
      }
      push(@newlinks, $link);
    }
    @links = @newlinks;
    if ($sort) {
      @links = sort(@links);
    }
    unshift (@links, $name);
    if ($empty || ($#links > 0)) {  # If only one item, list is empty.
      push(@found, join(' ', @links));
    }
  }
  return @found;
}

sub GetPageLinks {
  my ($name, $pagelink, $interlink, $urllink) = @_;
  my ($text, @links, %pg);

  @links = ();
  %pg = &GetCurrentPage($name);
  $text = $pg{'text'};
  $text =~ s/<html>((.|\n)*?)<\/html>/ /ig;
  $text =~ s/\<nowiki\\>(.|\n)*?\<\/nowiki\>/ /ig;
  $text =~ s/(^|\n)\&lt;pre\&gt;(.|\n)*?\n\&lt;\/pre\&gt;/ /ig;
  if ($interlink) {
    $text =~ s/''+/ /g;  # Quotes can adjacent to inter-site links
    $text =~ s/$InterLinkPattern/push(@links, &StripUrlPunct($1)), ' '/ge;
  } else {
    $text =~ s/$InterLinkPattern/ /g;
  }
  if ($urllink) {
    $text =~ s/''+/ /g;  # Quotes can adjacent to URLs
    $text =~ s/$UrlPattern/push(@links, &StripUrlPunct($1)), ' '/ge;
  } else {
    $text =~ s/$UrlPattern/ /g;
  }
  if ($pagelink) {
    $text =~ s/$LinkPattern/push(@links, &StripUrlPunct($1)), ' '/ge;
  }
  return @links;
}

sub ActionPost {
  my $string = &GetParam("text", undef);
  my $id = &GetParam("title", "");
  my $summary = &GetParam("summary", "");
  my $oldtime = &GetParam("oldtime", "");
  my $oldconflict = &GetParam("oldconflict", "");
  my $is_edit = 0;
  my $edit_time = time;
  my $authorAddr = $ENV{REMOTE_ADDR};
  my ($edit_diff, $old, $newAuthor, $preview);

  if (!$EditAllowed) {
    &ReportError("Editing not allowed: $SiteName is read-only.");
    return;
  }
  %page = &GetCurrentPage($id);
  $old = $page{text};

  $preview = 0;
  $preview = 1  if (&GetParam("Preview", "") ne "");
  if (!$preview && ($old eq $string)) {  # No changes (ok for preview)
    &ReBrowsePage($id, "");
    return;
  }
  $string =~ s/$FS//g;
  $summary =~ s/$FS//g;
  $summary =~ s/[\r\n]//g;
  if (($UserID > 399) || ($page{userid} > 399))  {
    $newAuthor = ($UserID ne $page{userid});       # known user(s)
  } else {
    $newAuthor = ($page{logaddr} ne $authorAddr);  # hostname fallback
  }
  if ($id eq "SampleUndefinedPage") {
    ReportError("SampleUndefinedPage cannot be defined.");
    return;
  }
  # Detect editing conflicts and resubmit edit
  if ($newAuthor && ($oldtime != $page{'timestamp'})) {
    if ($oldconflict>0) {  # Conflict again...
      &ActionEdit($id, 0, 2, $page{'timestamp'}, $string, $preview);
    } else {
      &ActionEdit($id, 0, 1, $page{'timestamp'}, $string, $preview);
    }
    return;
  }
  if ($preview) {
    &ActionEdit($id, 0, 0, $page{'timestamp'}, $string, 1);
    return;
  }

  &RequestLock();

  $edit_diff = "";
  if ($UseDiff) {
    $edit_diff = &GetDiff($old, $string);
    &WriteDiff($id, $edit_time, $edit_diff);
  }
  if (&GetParam("recent_edit", "") eq 'on') {
    $is_edit = 1;
  }
  my $user = &GetParam("username", "");
  # -- if the person doing editing chooses, send out email notification
  if ($UseEmailNotify) {
    EmailNotify($id, $user) if &GetParam("do_email_notify", "") eq 'on';
  }
  &SetCurrentPage($id, $string, $newAuthor, $is_edit, $edit_diff);
  &WriteRcLog($id, $summary, $is_edit, $edit_time, $UserID, $user);
  &ReleaseLock();
  &ReBrowsePage($id, "");
}

# Send an email message.
sub SendEmail {
  my ($to, $from, $reply, $subject, $message) = @_;
    ### debug
    ## print "Content-type: text/plain\n\n";
    ## print " to: '$to'\n";
    ## return;
  # sendmail options:
  #    -odq : send mail to queue (i.e. later when convenient)
  #    -oi  : do not wait for "." line to exit
  #    -t   : headers determine recipient.
  open (SENDMAIL, "| $SendMail -oi -t ") or die "Can't send email: $!\n";
  print SENDMAIL <<"EOF";
From: $from
To: $to
Reply-to: $reply
Subject: $subject\n
$message
EOF
  close(SENDMAIL) or warn "sendmail didn't close nicely";
}

## Email folks who want to know a note that a page has been modified. - JimM.
sub EmailNotify {
  local $/ = "\n";   # don't slurp whole files in this sub.
  if ($UseEmailNotify) {
    my ($id, $user) = @_;
    if ($user) {
      $user = " by $user";
    }
    my $address;
    open(EMAIL, "$DataDir/emails")
      or die "Can't open $DataDir/emails: $!\n";
    $address = join ",", <EMAIL>;
    $address =~ s/\n//g;
    close(EMAIL);
    my $home_url = $q->url();
    my $page_url = $home_url . "?$id";
    my $editors_summary = $q->param("summary");
    if (($editors_summary eq "*") or ($editors_summary eq "")){
      $editors_summary = "";
    }
    else {
      $editors_summary = "\n Summary: $editors_summary";
    }
    my $content = <<"END_MAIL_CONTENT";

 The $SiteName page $id at
   $page_url
 has been changed$user to version $FullPage{pagerev}. $editors_summary

 (Replying to this notification will
  send email to the entire mailing list,
  so only do that if you mean to.

  To remove yourself from this list, visit
  ${home_url}?action=editprefs .)
END_MAIL_CONTENT
    my $subject = "The $id page at $SiteName has been changed.";
    # I'm setting the "reply-to" field to be the same as the "to:" field
    # which seems appropriate for a mailing list, especially since the
    # $EmailFrom string needn't be a real email address.
    &SendEmail($address, $EmailFrom, $address, $subject, $content);
  }
}

sub ForceReleaseLock {
  my $n = 0;
  my $limit = 5;

  # First try to obtain lock (in case of normal edit lock)
  while (($n < $limit) && (mkdir($LockDir, 0555) == 0)) {
    $n++;
    sleep(3);
  }
  &ReleaseLock();  # Release the lock, even if we didn't get it.
  if ($n == $limit) {
    return 1;  # Lock was forced.
  }
  return 0;  # Lock was not forced.
}

# Note: call within lock.
sub SavePage {
  my ($id) = @_;
  my $file = &GetPageDirectory($id) . "$id.db";

  $FullPage{date} = &CalcDayNow;
  $FullPage{timestamp} = $^T;
  $FullPage{pagerev}++;

  my $string = join($FS1, %FullPage);
  &CreatePageDir(".", $id);
  &WriteStringToFile($file, $string);
  if ($UseCache) {
    UnlinkHtmlCache($id);          # Old cached copy is invalid
    if ($FullPage{pagerev} == 1) { # If this is a new page...
      &NewPageCacheClear($id);     # ...uncache pages linked to this one.
    }
  }
}

sub SetCurrentPage {
  my ($id, $newText, $newAuthor, $minorEdit, $diff) = @_;

  %page = &GetCurrentPage($id);
  &SaveHistory($id, $newText, $newAuthor, $minorEdit, $diff);

  $page{logaddr} = $ENV{REMOTE_ADDR};
  $page{logname} = &GetRemoteHost();
  $page{userid}  = $UserID;
  $page{text} = $newText;
  $page{date} = &CalcDayNow;
  $page{timestamp} = $^T;
  $page{revision}++;
  $FullPage{'current'} = join($FS3, %page);
  &SavePage($id);
}

sub SaveHistory {
  my ($id, $newText, $newAuthor, $minorEdit, $diff) = @_;
  my ($majorDiff, $authorDiff);

  $page{minordiff} = $diff;
  if ($minorEdit) {
    $page{majordiff} = &GetDiff($page{majorcopy}, $newText);
  } else {
    $page{majordiff} = "1";
    $page{majorcopy} = $page{text};
  }
  if ($page{revision} < 1) {
    $newAuthor = 1;  # First author is always a new author
  }
  if ($newAuthor) {
    $page{authordiff} = "1";
    $page{authorcopy} = $page{text};
  } else {
    $page{authordiff} = &GetDiff($page{authorcopy}, $newText);
  }
}

sub GetRemoteHost {
  my ($rhost, $iaddr);

  $rhost = $ENV{REMOTE_HOST};
  if ($rhost eq "") {
    # Catch errors (including bad input) without aborting the script
    eval '$iaddr = inet_aton($ENV{REMOTE_ADDR});'
         . '$rhost = gethostbyaddr($iaddr, AF_INET)';
  }
  if ($rhost eq "") {
    $rhost = $ENV{REMOTE_ADDR};
    $rhost =~ s/\d+$/xxx/;      # Be somewhat anonymous
  }
  return $rhost;
}

sub AllPagesList {
  my (@pages, @dirs, $id, $dir);

  @dirs = qw(A B C D E F G H I J K L M N O P Q R S T U V W X Y Z other);
  foreach $dir (@dirs) {
    while (<$dir/*.db $dir/*/*.db>) {
      m|^[^/]+/(\S*).db|;
      $id = $1;
      push(@pages, $id);
    }
  }
  return sort(@pages);
}

sub SearchTitleAndBody {
  my ($string) = @_;
  my ($name, @found);

  foreach $name (&AllPagesList()) {
    my %pg = &GetCurrentPage($name);
    if (($pg{text} =~ /$string/i) || ($name =~ /$string/i)) {
      push(@found, $name);
    }
  }
  return @found;
}

sub SearchBody {
  my ($string) = @_;
  my ($name, @found);

  foreach $name (&AllPagesList()) {
    my %pg = &GetCurrentPage($name);
    if ($pg{text} =~ /$string/i){
      push(@found, $name);
    }
  }
  return @found;
}

sub UnlinkHtmlCache {
  my ($id) = @_;
  my $id_file;

  $id_file = $HtmlDir . "/" . &GetPageDirectory($id) . "$id.htm";
  if (-f $id_file) {
    unlink($id_file);
  }
}

sub NewPageCacheClear {
  my ($id) = @_;
  my $name;

  return if (!$UseCache);
  $id =~ s|.+/|/|;  # If subpage, search for just the subpage
  foreach $name (&SearchBody($id)) {
    &UnlinkHtmlCache($name);
  }
}

sub DoUnlock {
  my $LockMessage = "Normal Unlock.";

  print &GetHeader("","Removing edit lock", "");
  print "<p>This operation may take several seconds...\n";
  if (&ForceReleaseLock()) {
    $LockMessage = "Forced Unlock.";
  }
  print "<br><h2>$LockMessage</h2>";
  &PrintCommonFooter();
}

# Note: all diff and recent-list operations should be done within locks.
sub WriteRcLog {
  my ($id, $summary, $is_edit, $edit_time, $uid, $name) = @_;
  my ($extraTemp, %extra);
  my $rhost = &GetRemoteHost();

  %extra = ();
  $extra{'id'} = $uid  if ($uid>0);
  $extra{'name'} = $name  if ($name ne "");
  $extraTemp = join($FS2, %extra);
  # The two fields at the end end of a line are kind and extension-hash
  my $rc_line = join($FS3, $edit_time, $id, $summary,
                     $is_edit, $rhost, "0", $extraTemp);
  if (!open(OUT, ">>$RcFile")) {
    die "RecentChanges log error: $!";
  }
  print OUT  $rc_line . "\n";
  close(OUT);
}

sub GetDiff {
  my ($old, $new) = @_;
  my $diff_out = "";

  &WriteStringToFile("old_diff", $old);
  &WriteStringToFile("new_diff", $new);

  $diff_out = `diff old_diff new_diff`;
  $diff_out =~ s/\\ No newline.*\n//g;   # Get rid of common complaint.
  # No need to unlink temp files--next diff will just overwrite.

  return $diff_out;
}

sub WriteDiff {
  my ($id, $edit_time, $diff_string) = @_;

  open (OUT, ">>diff_log") or die ("cant write diff_log");
  print OUT  "------\n" . $id . "|" . $edit_time . "\n";
  print OUT  $diff_string;
  close(OUT);
}
#END_OF_OTHER_CODE

&DoWikiRequest();   # Do everything.
# == End of UseModWiki script. ===========================================
